from collections.abc import Callable

import felis.order
from felis import monad
from felis.currying import curry
from felis.order import Order
from felis.predicate import Predicate

__all__ = ["identity", "map", "join", "bind", "compose", "then", "fold_left", "filter", "sort"]


def identity[T](value: T) -> list[T]:
    return [value]


@curry
def map[From, To](list_value: list[From], function: Callable[[From], To]) -> list[To]:
    return [function(value) for value in list_value]


def join[T](list_list_value: list[list[T]]) -> list[T]:
    return [value for list_value in list_list_value for value in list_value]


bind = monad.bind(map)(join)


compose = monad.compose(bind)


then = monad.then(bind)


@curry
@curry
def fold_left[T](list: list[T], empty: T, add: Callable[[T], Callable[[T], T]]) -> T:
    result = empty
    for value in list:
        result = add(value)(result)
    return result


@curry
def filter[T](list: list[T], predicate: Predicate[T]) -> list[T]:
    return [value for value in list if predicate(value)]


@curry
def sort[T](list: list[T], order: Order[T]) -> list[T]:
    return sorted(list, key=felis.order.rich_comparison(order))



from collections.abc import Callable

from felis import monad
from felis.currying import curry

__all__ = ["identity", "map", "inject", "join", "bind", "compose", "then"]


def identity[T](value: T) -> T:
    return value


@curry
def map[From, To](value: From, function: Callable[[From], To]) -> To:
    return function(value)


@curry
def inject[T, MT](m_value: MT, m_identity: Callable[[T], MT]) -> MT:
    return m_value


join = inject(identity)


bind = monad.bind(map)(join)


compose = monad.compose(bind)


then = monad.then(bind)



from typing import Protocol, Self

__all__ = ["SupportsRichComparison"]


class SupportsRichComparison(Protocol):
    def __lt__(self, other: Self, /) -> bool: ...

    def __gt__(self, other: Self, /) -> bool: ...

    def __le__(self, other: Self, /) -> bool: ...

    def __ge__(self, other: Self, /) -> bool: ...



from collections.abc import Callable

import felis.bool
from felis import function, identity

__all__ = ["Predicate", "both", "either", "negate", "same", "different"]


type Predicate[T] = Callable[[T], bool]


both = function.add(felis.bool.both)


either = function.add(felis.bool.either)


negate = identity.compose(felis.bool.negate)


same = function.add(felis.bool.same)


different = function.add(felis.bool.different)



from collections.abc import Callable
from dataclasses import dataclass

import felis.identity
from felis import monad
from felis.currying import curry

__all__ = ["Option", "Some", "identity", "map", "inject", "join", "bind", "compose", "then"]


type Option[T] = None | Some[T]


@dataclass(frozen=True)
class Some[T]:
    value: T


identity = Some


@curry
def map[From, To](option_value: Option[From], function: Callable[[From], To]) -> Option[To]:
    match option_value:
        case None:
            return None
        case Some(value):
            return Some(function(value))


@curry
def inject[T, MOptionT](option_m_option_value: Option[MOptionT], m_identity: Callable[[Option[T]], MOptionT]) -> MOptionT:
    match option_m_option_value:
        case None:
            return m_identity(None)
        case Some(m_option_value):
            return m_option_value


join = inject(felis.identity.identity)


bind = monad.bind(map)(join)


compose = monad.compose(bind)


then = monad.then(bind)



import felis.identity
from felis import coroutine, monad, option

__all__ = ["identity", "map", "join", "bind", "compose", "then"]


identity = felis.identity.compose(coroutine.identity)(option.identity)


map = felis.identity.compose(coroutine.map)(option.map)


join = coroutine.bind(option.inject(coroutine.identity))


bind = monad.bind(map)(join)


compose = monad.compose(bind)


then = monad.then(bind)



from collections.abc import Callable

from felis.currying import curry

__all__ = ["bind", "compose", "then"]


@curry
@curry
@curry
def bind[From, MFrom, MTo, MMTo](
    m_value: MFrom,
    function: Callable[[From], MTo],
    join: Callable[[MMTo], MTo],
    map: Callable[[Callable[[From], MTo]], Callable[[MFrom], MMTo]],
) -> MTo:
    return join(map(function)(m_value))


@curry
@curry
@curry
def compose[From, Intermediate, MIntermediate, MTo](
    value: From,
    first: Callable[[From], MIntermediate],
    second: Callable[[Intermediate], MTo],
    bind: Callable[[Callable[[Intermediate], MTo]], Callable[[MIntermediate], MTo]],
) -> MTo:
    return bind(second)(first(value))


@curry
@curry
def then[First, MFirst, MSecond](first: MFirst, second: MSecond, bind: Callable[[Callable[[First], MSecond]], Callable[[MFirst], MSecond]]) -> MSecond:
    return bind(lambda _: second)(first)



from collections.abc import Callable

__all__ = ["curry", "uncurry"]


def curry[First, *Rest, Result](function: Callable[[First, *Rest], Result]) -> Callable[[*Rest], Callable[[First], Result]]:
    return lambda *rest: lambda first: function(first, *rest)


def uncurry[First, *Rest, Result](function: Callable[[*Rest], Callable[[First], Result]]) -> Callable[[First, *Rest], Result]:
    return lambda first, *rest: function(*rest)(first)



from dataclasses import dataclass

from felis import predicate
from felis.currying import curry

__all__ = [
    "Ordering",
    "Worse",
    "Same",
    "Better",
    "worse",
    "same_or_better",
    "not_worse",
    "same",
    "different",
    "better",
    "same_or_worse",
    "not_better",
    "reverse",
    "neutral",
    "add",
]


type Ordering = Worse | Same | Better


@dataclass(frozen=True)
class Worse:
    pass


@dataclass(frozen=True)
class Same:
    pass


@dataclass(frozen=True)
class Better:
    pass


worse = curry(isinstance)(Worse)


same_or_better = not_worse = predicate.negate(worse)


same = curry(isinstance)(Same)


different = predicate.negate(same)


better = curry(isinstance)(Better)


same_or_worse = not_better = predicate.negate(better)


def reverse(ordering: Ordering) -> Ordering:
    match ordering:
        case Worse():
            return Better()
        case Same():
            return Same()
        case Better():
            return Worse()


neutral = Same()


@curry
def add(first: Ordering, second: Ordering) -> Ordering:
    match first:
        case Same():
            return second
        case _:
            return first



from collections.abc import Callable

import felis.identity
from felis import monad
from felis.currying import curry

__all__ = ["Function", "neutral", "neutral2", "add", "add2", "identity", "map", "map2", "comap", "bimap", "apply", "join", "bind", "compose", "then"]


type Function[From, To] = Callable[[From], To]


@curry
def neutral[M](_: object, m_neutral: M) -> M:
    return m_neutral


neutral2 = felis.identity.compose(neutral)(neutral)


@curry
@curry
@curry
def add[M, T](value: T, first: Function[T, M], second: Function[T, M], m_add: Callable[[M], Callable[[M], M]]) -> M:
    return m_add(second(value))(first(value))


add2 = felis.identity.compose(add)(add)


@curry
def identity[T](_: object, value: T) -> T:
    return value


@curry
@curry
def map[T, From, To](value: T, function_value: Function[T, From], function: Callable[[From], To]) -> To:
    return function(function_value(value))


map2 = felis.identity.compose(map)(map)


@curry
@curry
def comap[T, From, To](value: To, function_value: Function[From, T], function: Callable[[To], From]) -> T:
    return function_value(function(value))


@curry
@curry
@curry
def bimap[FromFrom, FromTo, ToFrom, ToTo](
    value: FromTo,
    function: Function[FromFrom, ToFrom],
    map: Callable[[ToFrom], ToTo],
    comap: Callable[[FromTo], FromFrom],
) -> ToTo:
    return map(function(comap(value)))


@curry
@curry
def apply[T, From, To](value: T, function_value: Function[T, From], function: Function[T, Callable[[From], To]]) -> To:
    return function(value)(function_value(value))


@curry
def join[From, To](value: From, function_function_value: Function[From, Function[From, To]]) -> To:
    return function_function_value(value)(value)


bind = monad.bind(map)(join)


compose = monad.compose(bind)


then = monad.then(bind)



import collections.abc
from collections.abc import Callable
from typing import Any

from felis import monad
from felis.currying import curry

__all__ = ["Coroutine", "identity", "map", "join", "bind", "compose", "then"]


type Coroutine[T] = collections.abc.Coroutine[Any, Any, T]


async def identity[T](value: T) -> T:
    return value


@curry
async def map[From, To](coroutine_value: Coroutine[From], function: Callable[[From], To]) -> To:
    return function(await coroutine_value)


async def join[T](coroutine_coroutine_value: Coroutine[Coroutine[T]]) -> T:
    return await (await coroutine_coroutine_value)


bind = monad.bind(map)(join)


compose = monad.compose(bind)


then = monad.then(bind)



import builtins
from collections.abc import Callable
from dataclasses import dataclass

from felis import monad
from felis.currying import curry
from felis.lazy import Lazy

__all__ = ["IO", "run", "identity", "map", "join", "bind", "compose", "then", "input", "print"]


@dataclass(frozen=True)
class IO[T]:
    value: Lazy[T]


def run[T](io_value: IO[T]) -> T:
    return io_value.value()


identity = IO


@curry
def map[From, To](io_value: IO[From], function: Callable[[From], To]) -> IO[To]:
    return IO(lambda: function(io_value.value()))


def join[T](io_io_value: IO[IO[T]]) -> IO[T]:
    return IO(lambda: io_io_value.value().value())


bind = monad.bind(map)(join)


compose = monad.compose(bind)


then = monad.then(bind)


input = IO(builtins.input)


def print(value: object) -> IO[None]:
    return IO(lambda: builtins.print(value))



from collections.abc import Callable

from felis import monad
from felis.currying import curry

__all__ = ["Lazy", "run", "identity", "map", "join", "bind", "compose", "then"]


type Lazy[T] = Callable[[], T]


def run[T](lazy_value: Lazy[T]) -> T:
    return lazy_value()


def identity[T](value: T) -> Lazy[T]:
    return lambda: value


@curry
def map[From, To](lazy_value: Lazy[From], function: Callable[[From], To]) -> Lazy[To]:
    return lambda: function(lazy_value())


def join[T](lazy_lazy_value: Lazy[Lazy[T]]) -> Lazy[T]:
    return lambda: lazy_lazy_value()()


bind = monad.bind(map)(join)


compose = monad.compose(bind)


then = monad.then(bind)



from collections.abc import Callable
from dataclasses import dataclass
from typing import Protocol

import felis.identity
from felis import monad
from felis.currying import curry

__all__ = ["Either", "Left", "Right", "identity", "map", "inject", "join", "bind", "compose", "then", "catch"]


type Either[L, R] = Left[L] | Right[R]


@dataclass(frozen=True)
class Left[T]:
    value: T


@dataclass(frozen=True)
class Right[T]:
    value: T


identity = Right


@curry
def map[L, From, To](either_value: Either[L, From], function: Callable[[From], To]) -> Either[L, To]:
    match either_value:
        case Left(value):
            return Left(value)
        case Right(value):
            return Right(function(value))


@curry
def inject[L, R, MEitherR](either_m_either_value: Either[L, MEitherR], m_identity: Callable[[Either[L, R]], MEitherR]) -> MEitherR:
    match either_m_either_value:
        case Left(value):
            return m_identity(Left(value))
        case Right(m_either_value):
            return m_either_value


# TODO: remove this
class Join(Protocol):
    @staticmethod
    def __call__[L, R](either_either_value: Either[L, Either[L, R]], /) -> Either[L, R]: ...


join: Join = inject(felis.identity.identity)


bind = monad.bind(map)(join)


compose = monad.compose(bind)


then = monad.then(bind)


@curry
@curry
def catch[E: BaseException, From, To](value: From, function: Callable[[From], To], exception_type: type[E]) -> Either[E, To]:
    try:
        return Right(function(value))
    except exception_type as exception:
        return Left(exception)



from collections.abc import Callable

from felis import monad
from felis.currying import curry
from felis.lazy import Lazy

__all__ = [
    "State",
    "ReversedState",
    "identity",
    "map",
    "join",
    "bind",
    "compose",
    "then",
    "reversed_join",
    "reversed_bind",
    "reversed_compose",
    "reversed_then",
]


type State[S, T] = Callable[[S], tuple[T, S]]


type ReversedState[S, T] = State[Lazy[S], T]


@curry
def identity[S, T](state: S, value: T) -> tuple[T, S]:
    return (value, state)


@curry
@curry
def map[S, From, To](state: S, state_value: State[S, From], function: Callable[[From], To]) -> tuple[To, S]:
    value, new_state = state_value(state)
    return function(value), new_state


@curry
def join[S, T](state: S, state_state_value: State[S, State[S, T]]) -> tuple[T, S]:
    state_value, state = state_state_value(state)
    value, new_state = state_value(state)
    return value, new_state


bind = monad.bind(map)(join)


compose = monad.compose(bind)


then = monad.then(bind)


@curry
def reversed_join[S, T](state: Lazy[S], reversed_state_reversed_state_value: ReversedState[S, ReversedState[S, T]]) -> tuple[T, Lazy[S]]:
    reversed_state_value, new_state = reversed_state_reversed_state_value(lambda: state())
    value, state = reversed_state_value(state)
    return value, new_state


reversed_bind = monad.bind(map)(reversed_join)


reversed_compose = monad.compose(reversed_bind)


reversed_then = monad.then(reversed_bind)



from collections.abc import Callable
from dataclasses import dataclass
from typing import Self

from felis import function, ordering
from felis.currying import curry
from felis.ordering import Ordering
from felis.typing import SupportsRichComparison

__all__ = [
    "Order",
    "worse",
    "same_or_better",
    "not_worse",
    "same",
    "different",
    "better",
    "same_or_worse",
    "not_better",
    "reverse",
    "neutral",
    "add",
    "map",
    "dunder",
    "rich_comparison",
]


type Order[T] = Callable[[T], Callable[[T], Ordering]]


worse = function.map2(ordering.worse)


same_or_better = not_worse = function.map2(ordering.not_worse)


same = function.map2(ordering.same)


different = function.map2(ordering.different)


better = function.map2(ordering.better)


same_or_worse = not_better = function.map2(ordering.not_better)


reverse = function.map2(ordering.reverse)


neutral = function.neutral2(ordering.neutral)


add = function.add2(ordering.add)


map = function.apply(function.bimap)(function.comap)


@curry
def dunder[T: SupportsRichComparison](first: T, second: T) -> Ordering:
    if first > second:
        return ordering.Better()
    if first < second:
        return ordering.Worse()
    return ordering.Same()


def rich_comparison[T](order: Order[T]) -> Callable[[T], SupportsRichComparison]:
    @dataclass(frozen=True)
    class RichComparison:
        value: T

        def __eq__(self, other: object, /) -> bool:
            if isinstance(other, RichComparison):
                return same(order)(other.value)(self.value)
            return NotImplemented

        def __ne__(self, other: object, /) -> bool:
            if isinstance(other, RichComparison):
                return different(order)(other.value)(self.value)
            return NotImplemented

        def __lt__(self, other: Self, /) -> bool:
            if isinstance(other, RichComparison):
                return worse(order)(other.value)(self.value)
            return NotImplemented

        def __gt__(self, other: Self, /) -> bool:
            if isinstance(other, RichComparison):
                return better(order)(other.value)(self.value)
            return NotImplemented

        def __le__(self, other: Self, /) -> bool:
            if isinstance(other, RichComparison):
                return same_or_worse(order)(other.value)(self.value)
            return NotImplemented

        def __ge__(self, other: Self, /) -> bool:
            if isinstance(other, RichComparison):
                return same_or_better(order)(other.value)(self.value)
            return NotImplemented

    return RichComparison



from felis import function
from felis.currying import curry

__all__ = ["both", "either", "negate", "same", "different"]


@curry
def both(first: bool, second: bool) -> bool:
    return first and second


@curry
def either(first: bool, second: bool) -> bool:
    return first or second


def negate(value: bool) -> bool:
    return not value


@curry
def same(first: bool, second: bool) -> bool:
    return first is second


different = function.map2(negate)(same)



from felis.coroutine import Coroutine
from felis.either import Either
from felis.function import Function
from felis.io import IO
from felis.lazy import Lazy
from felis.option import Option
from felis.order import Order
from felis.ordering import Ordering
from felis.predicate import Predicate
from felis.state import ReversedState, State

__all__ = ["Coroutine", "Either", "Function", "IO", "Lazy", "Option", "Order", "Ordering", "Predicate", "ReversedState", "State"]
